//---------------------------------------------------------------------------

#include <memory>
#include <strstream>
#include <vector>

#pragma hdrstop

#include "BITplayer.h"
#include "Chain.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

#define ELEMS(x) sizeof(x)/sizeof(x[0])
#define END(x) x + ELEMS(x)
#define RANGE(x) x,END(x)

using namespace std;

namespace JTAG
{

class BITPlayer::Implementation
	{
public:

	Implementation(Pod&	a_tm, Pod::TAPPort);

	void Play
		(std::istream&, ChainInfo*, StatusUpdate);

private:

	enum
		{
		BYPASS 	= 0x1F,
		IDCODE	= 0x09,
		CFG_IN	= 0x05,
		JSTART	= 0x0C
		};

	typedef unsigned char Byte;

	Pod&			tm;
	Pod::TAPPort	port;
	istream*		input;
	ChainInfo		chain;

	Pod::Device_Capabilities	caps;
	unsigned		input_length;
	StatusUpdate	status_update;

	void report_status(unsigned pos);

	// both the following start and end in IDLE
	void load_IR(Byte data);
	void load_DR(unsigned bits, Byte* data, bool correction = false);

	void determine_type();
	void deconfigure();
	void load_bitfile(unsigned bytes);
	void startup();
	};

#define IMPLEMENT BITPlayer::Implementation


IMPLEMENT::Implementation
	(Pod&	a_tm,
	 Pod::TAPPort	a_port)
	: 	tm(a_tm), port(a_port)
{
	tm.Get_General_Info(caps);
}

BITPlayer::BITPlayer
	(Pod& pod,
	 Pod::TAPPort	port)
	: impl(new Implementation(pod, port))
{}

BITPlayer* BITPlayer::new_BITPlayer
	(Pod&	pod,
	 Pod::TAPPort	port)
{ return new BITPlayer(pod, port); }

BITPlayer::~BITPlayer
	()
{}

void BITPlayer::Play
	(istream&		input,
	 ChainInfo*		chain,
	 StatusUpdate	update)
{
	impl->Play(input, chain, update);
}

void IMPLEMENT::report_status(unsigned pos)
{
	if (status_update)
		status_update(pos, input_length);
}

static unsigned get_number(istream& s, int digits)
{
	unsigned result = 0;
	while (digits--)
		{
		result = (result << 8) | s.get();
		if (s.fail() || s.eof()) throw Error_Exception("Unexpected EOF");
		}
	return result;
}

void IMPLEMENT::Play
	(istream&		a_input,
	 ChainInfo*		a_chain,
	 StatusUpdate	a_status)
{
	input = &a_input;
	status_update = a_status;
	if (a_chain)
		chain = *a_chain;
	else
		chain.IDCODE = 0;

	// get file size
	streampos start_pos = a_input.tellg();
	a_input.seekg(0, ios_base::end);
	input_length = a_input.tellg() - start_pos;

	// skip over bitfile headers
	for (int skip1 = 3; skip1--;)
		{
		a_input.seekg(start_pos);
		start_pos += int(2 + get_number(a_input, 2));
		}

	for (;;)
		{
		a_input.seekg(start_pos);
		if (a_input.get() == 'e') break;
		start_pos += int(3 + get_number(a_input, 2));
		}

	// get length (in bytes)
	a_input.seekg(start_pos + 1);
	unsigned bitlength = get_number(a_input, 4);

	// init the TAP - RESET IDLE
	tm.Run_TAP_X(port, 6, 0x01F);

	if (!chain.IDCODE)
		determine_type();

	// do it
	deconfigure();
	load_bitfile(bitlength);
	startup();

	tm.Flush_Commands();
}

void IMPLEMENT::load_IR(Byte data)	// fixed at 5 bits
{
	// SELECTDR SELECTIR CAPTUREIR SHIFTIR
	tm.Run_TAP_X(port, 4, 0x003);

	if (chain.IR_bits_before)
		{ tm.Do_Shift(port, chain.IR_bits_before, false, NULL); }

	bool at_end = !chain.IR_bits_after;
	tm.Do_Shift(port, 5, at_end, &data);
	if (!at_end)
		{ tm.Do_Shift(port, chain.IR_bits_after, true, NULL); }

	// UPDATEIR IDLE
	tm.Run_TAP_X(port, 2, 0x001);
}

void IMPLEMENT::load_DR(unsigned bits, Byte* data, bool correction)
{
	// SELECTDR CAPTUREDR SHIFTDR
	tm.Run_TAP_X(port, 3, 0x001);

	if (chain.devices_before)
		{
		unsigned shift = correction ?
			(32 - chain.devices_before) : chain.devices_before;
		// 32 - devices before is from XAPP188
		tm.Do_Shift(port, shift, false, NULL);
		}

	bool at_end = !chain.devices_after;
	tm.Do_Shift(port, bits, at_end, data);
	if (!at_end)
		{ tm.Do_Shift(port, chain.devices_after, true, NULL); }

	// UPDATEDR IDLE
	tm.Run_TAP_X(port, 2, 0x001);
}

void IMPLEMENT::deconfigure()
{
	static Byte Shutdown1[] =
		{ 0x0C, 0x80, 0x04, 0x80, 0x00, 0x05, 0xfd, 0xbc, 0x0c, 0x00,
		  0x01, 0x80, 0x00, 0x00, 0x00, 0xa0, 0x0c, 0x00, 0x01, 0x80,
		  0x00, 0x00, 0x00, 0xe0, 0x00, 0x00, 0x00, 0x00 };

	static Byte Shutdown2[] =
		{ 0x0c, 0x00, 0x01, 0x80, 0x00, 0x00, 0x00, 0x10, 0x0c, 0x80,
		  0x04, 0x80, 0x00, 0x05, 0xfc, 0xff, 0x0c, 0x00, 0x01, 0x80,
		  0x00, 0x00, 0x00, 0xa0, 0x0c, 0x00, 0x01, 0x80, 0x00, 0x00,
		  0x00, 0xe0, 0x00, 0x00, 0x00, 0x00 };

	static Byte Shutdown3[] =
		{ 0x0C, 0x80, 0x04, 0x80, 0x00, 0x05, 0xfd, 0xb4, 0x0c, 0x00,
		  0x01, 0x80, 0x00, 0x00, 0x00, 0xa0, 0x0c, 0x00, 0x01, 0x80,
		  0x00, 0x00, 0x00, 0xe0, 0x00, 0x00, 0x00, 0x00 };


	load_IR(BYPASS);

	load_IR(CFG_IN);
	load_DR(224, Shutdown1, true);
	load_IR(JSTART);
	load_DR(13, NULL);

	load_IR(CFG_IN);
	load_DR(288, Shutdown2, true);
	load_IR(JSTART);
	load_DR(13, NULL);

	load_IR(CFG_IN);
	load_DR(224, Shutdown3, true);
	load_IR(JSTART);
	load_DR(13, NULL);
}

static unsigned char flip[256];
static bool inited = false;

void IMPLEMENT::load_bitfile(unsigned bytes)
{

	if (!inited)
		{
		inited = true;
		/* build a bit-reversing array.  This is required because
		   the FPGA expects the bitfile with b0 as msb. */
		for (int n = 0; n < 256; ++n)
			{
			unsigned j = n;
			j = ((j & 0xAA) >> 1) | ((j & 0x55) << 1);
			j = ((j & 0xCC) >> 2) | ((j & 0x33) << 2);
			j = ((j & 0xF0) >> 4) | ((j & 0x0F) << 4);
			flip[n] = Byte(j);
			}
		}

	load_IR(CFG_IN);

	// SELECTDR CAPTUREDR SHIFTDR
	tm.Run_TAP_X(port, 3, 0x001);

	if (chain.devices_before)
		{
		// 32 - devices before is from XAPP188
		tm.Do_Shift(port, 32 - chain.devices_before, false, NULL);
		}

	bool at_end = !chain.devices_after;

	unsigned buf_len = (caps.buffer_bits / 8);
	vector<Byte> buf(buf_len);
	while (bytes)
		{
		unsigned chunk = min(bytes, buf_len);
		input->read(buf.begin(), chunk);
		if (input->fail())
			throw Error_Exception("could not read data");
		bytes -= chunk;

		vector<Byte>::iterator p = buf.begin();
		for (int n = chunk; n--;)
			{ *p++ = flip[*p]; }

		tm.Do_Shift(port, chunk * 8, !bytes && at_end, buf.begin());
		report_status(input->tellg());
		}

	if (!at_end)
		{ tm.Do_Shift(port, chain.devices_after, true, NULL); }

	// UPDATEDR IDLE
	tm.Run_TAP_X(port, 2, 0x003);
}

void IMPLEMENT::startup()
{
	load_IR(JSTART);
	load_DR(13, NULL);

	load_IR(BYPASS);

	// check DONE pin
	///!!!!
}

void IMPLEMENT::determine_type()
{
	if (!chain.IDCODE)
		{
		Chain	temp;
		temp.ReadChain(tm, port);
		if (temp.devices.size() > 1)
			throw Error_Exception("More than 1 device - explicit chain reqd");
		temp.GetChainInfo(0, chain);
		}

	switch ((chain.IDCODE >> 1) & 0x7FF)
		{
	case 0x49:	// Xilinx
		switch ((IDCODE >> 21) & 0x7F)
			{
		//case 0x02:	//	XCS
		case 0x03:		//XC2S
		//case 0x05:	//XC2SE
		//case 0x0A:	//XC3S

		//case 0x03:	//XCV
		//case 0x06:	//XCVE
		//case 0x08:	//XC2V
		//case 0x09:	//XC2VP

		//case 0x28:	/XC18V

		// algorithm = Spartan II
			return;
			}
		}

	throw Error_Exception("No programming alg. for device");
}

}

