//---------------------------------------------------------------------------

//#include <memory>
//#include <strstream>
//#include <vector>
#include <functional>

#pragma hdrstop

#include "Pod.h"
#include "Chain.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

using namespace std;

namespace JTAG
{

ChainInfo::ChainInfo(unsigned a_IDCODE)
	:	IDCODE(a_IDCODE),
		IR_bits(0),
		devices_before(0),
		IR_bits_before(0),
		devices_after(0),
		IR_bits_after(0)
{}

//	unsigned		IDCODE;
//	unsigned		IDCODE_mask;
//	unsigned		IR_bits;
//	unsigned		IDCODE_instruction;
//	string		dev_name;
//	string		dev_description;
//	string		BSDL_file;

Device::Device(unsigned a_IDCODE)
	:	IDCODE(a_IDCODE),
		IDCODE_mask(0xFFFFFFFE),
		IR_bits(0),
		IDCODE_instruction(0)
{}

void Device::Store(ostream& os)
{
	os	<< dev_name << endl;
	os	<< IR_bits << ','
		<< IDCODE_instruction << ','
		<< max_freq << ':'
		<< hex << IDCODE << '/'
		<< hex << IDCODE_mask << endl;
	os	<< dev_description << endl;
	os	<< BSDL_file << endl;
}

static istream& operator>>(istream& is, const char match)
{
	char c;
	is >> c;
	if (c != match)
		is.setstate(ios_base::failbit);
	return is;
}

void Device::Load(istream& is)
{
	getline(is, dev_name);
	is	>> IR_bits >> ','
		>> IDCODE_instruction >> ','
		>> max_freq >> ':'
		>> hex >> IDCODE >> '/'
		>> hex >> IDCODE_mask;
	//is.ignore(1);
	getline(is, dev_description);
	getline(is, BSDL_file);
}

//	vector<Device>		devices;

inline bool operator != (Device& a, unsigned b)
{ return ((a.IDCODE ^ b) & a.IDCODE_mask) == 0; }

Device* Device_Database::Find(unsigned IDCODE)
{
	Device* p =
		find(devices.begin(), devices.end(), IDCODE);
	return (p == devices.end()) ? 0 : p;
}

Device* Device_Database::Create(unsigned IDCODE, unsigned IR_bits)
{
	devices.resize(devices.size() + 1);
	Device* p = devices.end() - 1;
	p->IDCODE = IDCODE;
	p->IDCODE_mask = unsigned(-1);
	p->IR_bits = IR_bits;
	p->IDCODE_instruction = 0;
	p->dev_name = "Unknown";
	p->max_freq = 10e6;
	return p;
}

void Device_Database::Store(ostream& os)
{
	os << devices.size() << endl;
	for (vector<Device>::iterator p = devices.begin(); p != devices.end(); ++p)
		(*p).Store(os);
}

void Device_Database::Load(istream& is)
{
	unsigned size;
	is >> size;
	devices.clear();
	devices.resize(size);
	for (vector<Device>::iterator p = devices.begin(); p != devices.end(); ++p)
		(*p).Load(is);
}

//	vector<Device*>	devices;
//	vector<string>	data_files;

bool Chain::ValidateChain(Pod& tm, unsigned port)
{
	bool valid = false;

	// RESET IDLE SELECTDR CAPTUREDR SHIFTDR
	tm.Run_TAP_X(port, 9, 0x05F);

	unsigned pos;
	for (pos = 0;; ++pos)
		{
		Byte ID[4];
		tm.Do_Shift(port, 32, false, NULL, ID);
		IDCODE_t IDCODE = 0;
		for (int n = 4; n--;)
			IDCODE = (IDCODE << 8) | ID[n];

		if ((IDCODE == 0xFFFFFFFF) || (!(IDCODE & 1)))
			{
			if (pos == devices.size())
				valid = true;
			break;
			}

		if ((pos >= devices.size()) || (devices[pos]->IDCODE != IDCODE))
			break;
		}

	// EXIT1DR UPDATEDR IDLE
	tm.Run_TAP_X(port, 3, 0x003);

	return valid;
}

bool Chain::ReadChain(Pod& tm, unsigned port, Device_Database* gdb)
{
	devices.clear();
	data_files.clear();

	// RESET IDLE SELECTDR CAPTUREDR SHIFTDR
	tm.Run_TAP_X(port, 9, 0x05F);

	for (int max = 32; max--;)
		{
		Byte ID[4];
		tm.Do_Shift(port, 32, false, NULL, ID);
		IDCODE_t IDCODE = 0;
		for (int n = 4; n--;)
			IDCODE = (IDCODE << 8) | ID[n];
		if ((IDCODE == 0xFFFFFFFF) || (!(IDCODE & 1)))
			break;

		Device* p = 0;
		if (gdb)
			p = gdb->Find(IDCODE);

		IDCODEs.push_back(IDCODE);
		devices.push_back(p);
		data_files.push_back(string());
		}

	// EXIT1DR UPDATEDR IDLE
	tm.Run_TAP_X(port, 3, 0x003);
	return (devices.size() > 0);
}

bool Chain::ResolveDevices(Device_Database* gdb, ResolveCB f)
{
	for (unsigned p = 0; p < devices.size(); ++p)
		if (!devices[p])
			{
			Device* match = 0;
			if (gdb)
				match = gdb->Find(IDCODEs[p]);
			if (!match)
				{
				if (!f)
					return false;
				match =  f(*this, p);
				if (!match)
					return false;
				}
			devices[p] = match;
			}
	return true;
}

void Chain::GetChainInfo(unsigned position, ChainInfo& ci)
{
	ci.devices_before = position;
	ci.devices_after  = devices.size() - position - 1;
	ci.IR_bits_before =
	ci.IR_bits_after = 0;

	unsigned* accum = &ci.IR_bits_before;
	for (unsigned n = 0; n < devices.size(); ++n)
		{
		if (!devices[n])
			throw Error_Exception("Unknown device in chain");
			
		if (n == position)
			{
			ci.IR_bits = devices[n]->IR_bits;
			ci.IDCODE = devices[n]->IDCODE;
			accum = &ci.IR_bits_after;
			}
		else
			*accum += devices[n]->IR_bits;
		}
}

void Chain::Store(ostream& os)
{
	vector<Device*>::iterator	dev;
	vector<string>::iterator		file = data_files.begin();
	os << devices.size();
	for (dev = devices.begin(); dev != devices.end(); ++dev, ++file)
		os << hex << (*dev)->IDCODE << ':' << *file << endl;
}

void Chain::Load(istream& is)
{
	unsigned size;
	is >> size;
	IDCODEs.clear();
	devices.clear();
	data_files.clear();
	devices.reserve(size);
	data_files.reserve(size);
	IDCODEs.reserve(size);
	while (size--)
		{
		IDCODE_t IDCODE;
		string df;

		is >> hex >> IDCODE >> ':';
		getline(is, df);

		IDCODEs.push_back(IDCODE);
		devices.push_back(0);
		data_files.push_back(df);
		}
}

}

