
#if !defined(TAPMGR_H)
#define TAPMGR_H

#include <stdexcept>

namespace JTAG
{

#if defined(__WIN32__) || defined(_WIN32)
  #if defined(__DLL__)
	#define LINKAGE __declspec(dllexport)
  #else
	#define LINKAGE __declspec(dllimport)
	#pragma link "JTAGPod.lib"
  #endif
#else
  #define LINKAGE
#endif

typedef unsigned char Byte;

class LINKAGE Error_Exception
	: public std::runtime_error
	{
  public:
	explicit Error_Exception(const std::string&  what_arg )
		: std::runtime_error(what_arg) {}
	};

class LINKAGE Pod
	{

	// N.B. This class is not thread-safe.  If calling from multiple
	// threads, e.g. running two TAPs simultaneously, a mutex should
	// be used.

public:

	struct Device_Capabilities
		{
		const char*	ID_string;
		unsigned	num_TAPs;
		unsigned	num_buffers;
		unsigned	buffer_bits;
		unsigned	num_LEDs;
		unsigned	num_TCK_rates;
		const float* TCK_rates;
		unsigned	TMS_bits;
		};

	enum GPIO { isZ = 0, is0 = 1, is1 = 3 };

	struct TAP_Capabilities
		{
		unsigned	num_GPIOs;
		bool		canTristate;
		bool		hasTCKTerm;
		bool		hasTDOTerm;
		};

	struct TAP_Settings
		{
		bool		tristated;
		bool		ready;
		unsigned	current_TDO_buffer;
		unsigned	TCK_rate_index;
		enum TCK_Terminations { disabled, term_33, term_47, term_20 }
					TCK_term;
		enum TDO_Terminations { unterminated, passive_term }
					TDO_term;
		const GPIO*	Current_GPIO_Setting;
		};

	enum TAP_State
		{
		RESET, IDLE,
		DRSELECT, DRCAPTURE, DRSHIFT, DREXIT1, DRPAUSE, DREXIT2, DRUPDATE,
		IRSELECT, IRCAPTURE, IRSHIFT, IREXIT1, IRPAUSE, IREXIT2, IRUPDATE
		};

	virtual ~Pod();

	typedef unsigned TAPPort;

#define TAPMGRIF(s) virtual void s =0;

	TAPMGRIF( Get_General_Info(Device_Capabilities&) )
	TAPMGRIF( Set_LED(unsigned, bool) )

	TAPMGRIF( Flush_Commands() )

	TAPMGRIF( Get_TAP_Info(TAPPort, TAP_Capabilities&) )
	TAPMGRIF( Get_TAP_Settings(TAPPort, TAP_Settings&) )
	TAPMGRIF( Set_TAP_Settings(TAPPort, TAP_Settings&) )

	TAPMGRIF( Get_Current_State(TAPPort, TAP_State&) )
	TAPMGRIF( Do_State_Transition(TAPPort, TAP_State = RESET) )
	TAPMGRIF( Do_State_Move(TAPPort, int = 1 /* 0 or 1 */) )
	TAPMGRIF( Do_Shift(TAPPort, unsigned bits, bool end_of_shift,
					   Byte* TDO, Byte* TDI = 0) )
	TAPMGRIF( Do_Run_TCK(TAPPort, bool onoff) )


	TAPMGRIF( Write_TDO_Buffer(TAPPort, unsigned offset, unsigned bytes, Byte*) )
	TAPMGRIF( Read_TDI_Buffer(TAPPort, unsigned offset, unsigned bytes, Byte*) )
	TAPMGRIF( Copy_TDI_Buffer_To_TDO(TAPPort, unsigned offset, unsigned bytes) )
	TAPMGRIF( Run_TAP_X(TAPPort, unsigned bits, unsigned TMS, unsigned flags = 0) )

	TAPMGRIF( Set_TRST(TAPPort, bool) )
	TAPMGRIF( Pulse_TRST(TAPPort, unsigned ms) )

	TAPMGRIF( Set_RST(TAPPort, bool) )
	TAPMGRIF( Pulse_RST(TAPPort, unsigned ms) )

	TAPMGRIF( Set_DINT(TAPPort, bool) )
	TAPMGRIF( Pulse_DINT(TAPPort, unsigned ms) )

	TAPMGRIF( Get_TAP_GPIO_Input(TAPPort, unsigned, GPIO&) )
	TAPMGRIF( Set_TAP_GPIO(TAPPort, unsigned, GPIO) )
	TAPMGRIF( Pulse_TAP_GPIO(TAPPort, unsigned, unsigned ns, GPIO pulse, GPIO after) )

#undef TAPMGRIF

protected:

	Pod(){}
	};

enum USB_Pod_FPGA_Types
	{ XC2S30, XC2S50, XC2S100 };

LINKAGE void ConfigureHW(unsigned device, unsigned HW_rev, USB_Pod_FPGA_Types type);
LINKAGE void CountDevices(unsigned&);
LINKAGE Pod* new_Pod(unsigned device);

#undef LINKAGE

}

#endif

