//---------------------------------------------------------------------------

#include <memory>
#include <locale>
#include <strstream>

#pragma hdrstop

#include "SVFplayer.h"
#include "SimpleLexer.h"
#include "BitString.h"
#include "Chain.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)

#define ELEMS(x) sizeof(x)/sizeof(x[0])
#define END(x) x + ELEMS(x)
#define RANGE(x) x,END(x)

// Xilinx SVF files seem to assume this frequency for TCK in RUNTEST
const float XILINX_DEFAULT_TCK = 250e3;

using namespace std;

namespace JTAG
{

class SVFPlayer::Implementation
	{
public:

	Implementation(Pod&	a_tm, Pod::TAPPort);

	void Play
		(std::istream&, ChainInfo*, StatusUpdate, bool, bool);

private:

	enum Token
		{
		// the first 16 tokens are also the state codes, and must match
		// the declaration in Pod
		kRESET, kIDLE,
		kDRSELECT, kDRCAPTURE, kDRSHIFT, kDREXIT1, kDRPAUSE, kDREXIT2, kDRUPDATE,
		kIRSELECT, kIRCAPTURE, kIRSHIFT, kIREXIT1, kIRPAUSE, kIREXIT2, kIRUPDATE,

		// these tokens are the remaining syntactical elements of the SVF file
		kTerm, kComment1, kComment2, kStartHex, kStopHex,
		kDot, kMinus,
		kENDDR, kENDIR,
		kFREQUENCY, kHZ,
		kHDR, kHIR, kTDI, kTDO, kMASK, kSMASK,
		kPIO, kH, kL, kZ, kU, kD, kX,
		kPIOMAP, kIN, kINOUT, kOUT,
		kRUNTEST, kSEC, kMAXIMUM, kENDSTATE, kTCK, kSCK,
		kSDR, kSIR, kSTATE, kTDR, kTIR, kTRST, kON, kOFF, kABSENT,
		k_End
		};

	static struct PathTransition
		{
		Token	from;
		Token	to;
		unsigned char bits;
		unsigned char TMS;
		} std_transitions[];

	enum BitLists
		{ IDX_HIR, IDX_IR, IDX_TIR,
		  IDX_HDR, IDX_DR, IDX_TDR, CNT_HDT };

	struct RegisterData
		{
		Token		end_state;
		unsigned 	bits;
		BitString	TDI, TDO, MASK, SMASK;

		RegisterData() : bits(0) {}
		void reset();
		};

	RegisterData	reg[CNT_HDT];

	typedef void (Implementation::*Handler)(Token);

	Pod&	tm;
	Pod::TAPPort	port;
	auto_ptr<SimpleLexer>	lex;
	istream*		input;
	ChainInfo*		chain;
	bool			Xilinx_timing;
	bool			check_TDO;

	Pod::Device_Capabilities caps;
	StatusUpdate	status_update;

	int				initial_frequency, cur_frequency;
	unsigned		input_length;
	Token			runtest_run_state, runtest_end_state;

	class Action;
	static Action action_list[];

	void reset_TAP();
	void set_up_chain();
	void set_up_chained_reg(RegisterData&, unsigned, bool);

	void report_status(unsigned pos);

	void skipws();
	SimpleLexer::LexerResult Peek();
	SimpleLexer::LexerResult Next();
	Token PeekToken();
	Token NextToken();

	Token Accept(Token* begin, Token* end);
	void check_term();

	void parse_register(RegisterData&);
	void parse_bits(unsigned max_bits, BitString&);
	void parse_real(double&);

	void to_stable_state(Token);

	void do_TRST(Token);
	void do_end_state(Token);
	void do_header_trailer(Token);
	void do_scan(Token);
	void do_state(Token);
	void do_frequency(Token);
	void do_runtest(Token);
	};

#define IMPLEMENT SVFPlayer::Implementation

IMPLEMENT::PathTransition IMPLEMENT::std_transitions[] =
	{
	{ kRESET, kRESET,		1, 0x01 },
	{ kRESET, kIDLE,		1, 0x00 },
	{ kRESET, kDRPAUSE,		5, 0x0A },
	{ kRESET, kIRPAUSE,		6, 0x16 },
	{ kIDLE, kRESET,		3, 0x07 },
	{ kIDLE, kIDLE,			1, 0x00 },
	{ kIDLE, kDRPAUSE,		5, 0x05 },
	{ kIDLE, kIRPAUSE,		5, 0x0B },
	{ kDRPAUSE, kRESET,		5, 0x1F },
	{ kDRPAUSE, kIDLE,		3, 0x03 },
	{ kDRPAUSE, kDRPAUSE,	6, 0x17 },
	{ kDRPAUSE, kIRPAUSE,	7, 0x2F },
	{ kIRPAUSE, kRESET,		5, 0x1F },
	{ kIRPAUSE, kIDLE,		3, 0x03 },
	{ kIRPAUSE, kDRPAUSE,	5, 0x0F },
	{ kIRPAUSE, kIRPAUSE,	7, 0x2F }
	};

static const SimpleLexer::Keyword token_strings[] =
	{
	"RESET", "IDLE",
	"DRSELECT", "DRCAPTURE", "DRSHIFT", "DREXIT1", "DRPAUSE", "DREXIT2", "DRUPDATE",
	"IRSELECT", "IRCAPTURE", "IRSHIFT", "IREXIT1", "IRPAUSE", "IREXIT2", "IRUPDATE",
	";", "!", "//", "(", ")",
	".", "-",
	"ENDDR", "ENDIR",
	"FREQUENCY", "HZ",
	"HDR", "HIR", "TDI", "TDO", "MASK", "SMASK",
	"PIO", "H", "L", "Z", "U", "D", "X",
	"PIOMAP", "IN", "INOUT", "OUT",
	"RUNTEST", "SEC", "MAXIMUM", "ENDSTATE", "TCK", "SCK",
	"SDR", "SIR", "STATE", "TDR", "TIR", "TRST", "ON", "OFF", "ABSENT"
	};

class IMPLEMENT::Action
	: public std::pair<Token, Handler>
	{
public:
	Action(Token a, Handler b)
		: pair<Token, Handler>(a, b) {}
	bool operator != (Token x) const
		{ return x != first; }
	void execute(IMPLEMENT* obj, Token tok)
		{ (obj->*second)(tok); }
	};

IMPLEMENT::Action IMPLEMENT::action_list[] =
	{
	Action(kTRST, do_TRST),
	Action(kENDDR, 	do_end_state),
	Action(kENDIR, 	do_end_state),
	Action(kSTATE,	do_state),
	Action(kHIR,	do_header_trailer),
	Action(kHDR,	do_header_trailer),
	Action(kTIR,	do_header_trailer),
	Action(kTDR,	do_header_trailer),
	Action(kSIR,	do_scan),
	Action(kSDR,	do_scan),
	Action(kFREQUENCY, do_frequency),
	Action(kRUNTEST, do_runtest)
	};

void error(const string& msg)
{
	throw Error_Exception(msg);
}

IMPLEMENT::Implementation
	(Pod&	a_tm,
	 Pod::TAPPort	a_port)
	: 	tm(a_tm), port(a_port),
		lex(new SimpleLexer(RANGE(token_strings)))
{
	tm.Get_General_Info(caps);
}

SVFPlayer::SVFPlayer
	(Pod& pod,
	 Pod::TAPPort	port)
	: impl(new Implementation(pod, port))
{}

SVFPlayer* SVFPlayer::new_SVFPlayer
	(Pod&	pod,
	 Pod::TAPPort	port)
{ return new SVFPlayer(pod, port); }

SVFPlayer::~SVFPlayer
	()
{}

void IMPLEMENT::reset_TAP()
{
	tm.Run_TAP_X(port, 5, 0x0FF);
}

void IMPLEMENT::RegisterData::reset()
{
	TDI.reset();
	TDO.reset();
	MASK.reset();
	SMASK.reset();
	bits = 0;
}

void IMPLEMENT::set_up_chained_reg(RegisterData& b, unsigned bits, bool ones)
{
	b.reset();
	b.bits = bits;
	b.TDI.present = true;
	b.TDI.data.resize((bits / BitString::MAX_SHIFT) + 1, unsigned(ones ? -1 : 0));
}

void IMPLEMENT::set_up_chain()
{
	if (chain)
		{
		set_up_chained_reg(reg[IDX_HIR], chain->IR_bits_before, true);
		set_up_chained_reg(reg[IDX_HDR], chain->devices_before, false);
		set_up_chained_reg(reg[IDX_TIR], chain->IR_bits_after, true);
		set_up_chained_reg(reg[IDX_TDR], chain->devices_after, false);
		}
}

void IMPLEMENT::skipws()
{
	for (;;)
		switch (lex->Peek())
			{
		case SimpleLexer::rWhite:
		case SimpleLexer::rEOL:
			lex->Next();
			break;

		case SimpleLexer::rKeyword:
			if ((lex->index != kComment1) &&
				(lex->index != kComment2))
				return;

			do {} while (lex->Next() != SimpleLexer::rEOL);
			break;

		default:
			return;
			}
}

SimpleLexer::LexerResult IMPLEMENT::Peek()
{
	skipws();
	return lex->Peek();
}

SimpleLexer::LexerResult IMPLEMENT::Next()
{
	skipws();
	return lex->Next();
}

IMPLEMENT::Token IMPLEMENT::PeekToken()
{
	if (Peek() != SimpleLexer::rKeyword) return k_End;
	return Token(lex->index);
}

IMPLEMENT::Token IMPLEMENT::NextToken()
{
	if (Next() != SimpleLexer::rKeyword) return k_End;
	return Token(lex->index);
}

void SVFPlayer::Play
	(istream&		input,
	 ChainInfo*		chain,
	 StatusUpdate	update,
	 bool			Xilinx_timing,
	 bool				check_TDO)
{
	impl->Play(input, chain, update, Xilinx_timing, check_TDO);
}

void IMPLEMENT::report_status(unsigned pos)
{
	if (status_update)
		status_update(pos, input_length);
}

void IMPLEMENT::Play
	(istream&		a_input,
	 ChainInfo*		a_chain,
	 StatusUpdate	a_status,
	 bool			a_Xilinx_timing,
	 bool			a_check_TDO)
{
	try
		{
		// init members
		reg[IDX_DR].end_state = kIDLE;
		reg[IDX_IR].end_state = kIDLE;
		runtest_run_state = kIDLE;
		runtest_end_state = kIDLE;

		input = &a_input;
		status_update = a_status;
		chain = a_chain;
		Xilinx_timing = a_Xilinx_timing;
		check_TDO = a_check_TDO;
		set_up_chain();

		// get starting FREQUENCY
		Pod::TAP_Settings cur_ts;
		tm.Get_TAP_Settings(port, cur_ts);
		cur_frequency = initial_frequency = cur_ts.TCK_rate_index;

		// init the TAP
		reset_TAP();

		// get file size
		streampos start_pos = a_input.tellg();
		a_input.seekg(0, ios_base::end);
		input_length = a_input.tellg() - start_pos;
		a_input.seekg(start_pos, ios_base::beg);

		// and parse/execute the file
		lex->Open(a_input);

		for (bool run = true; run;)
			{
			report_status(lex->tellg());

			switch (Next())
				{
			case SimpleLexer::rKeyword:
				{
				Action* action =
					std::find(RANGE(action_list), Token(lex->index));
				if (action == END(action_list))
					error("Unrecognized SVF command");
				action->execute(this, Token(lex->index));
				}
				break;

			case SimpleLexer::rEOF:
				run = false;
				break;

			default:
				error("SVF Syntax error");
				break;
				}
			}

        // clock speed back to normal
		cur_ts.TCK_rate_index = initial_frequency;
		tm.Set_TAP_Settings(port, cur_ts);

		tm.Flush_Commands();
		}

	catch (const Error_Exception& msg)
		{
		ostrstream os;

		os << "Error on line " << lex->line << " near column " << lex->column
		   << ":" << endl << msg.what();
		error(os.str());
		}
}

IMPLEMENT::Token IMPLEMENT::Accept(Token* begin, Token* end)
{
	if (Next() == SimpleLexer::rKeyword)
		{
		for (Token* p = begin; p != end; ++p)
			if (*p == lex->index)
				return (*p);
		}

	// format error message
	string msg("Expected one of: ");
	for (; begin != end;)
		{
		msg += token_strings[*begin];
		if (++begin != end)
			msg += ", ";
		}

	error(msg);
	return Token(-1);
}

void IMPLEMENT::check_term()
{
	if (NextToken() != kTerm)
		error("expect end of command");
}

void IMPLEMENT::do_TRST(Token)
{
	Token ok[] = { kON, kOFF, kZ, kABSENT };
	switch (Accept(RANGE(ok)))
		{
	case kABSENT:
	case kZ:	// can't control this separately
	case kOFF:
		tm.Set_TRST(port, false);
		break;

	case kON:
		tm.Set_TRST(port, true);
		break;
		}

	check_term();
}

void IMPLEMENT::do_end_state(Token token)
{
	Token value, ok[] = { kIRPAUSE, kDRPAUSE, kRESET, kIDLE };
	value = Accept(RANGE(ok));
	if (token == kENDDR)
		reg[IDX_DR].end_state = value;
	else
		reg[IDX_IR].end_state = value;
	check_term();
}

void IMPLEMENT::to_stable_state(Token tok)
{
	Pod::TAP_State cur;
	tm.Get_Current_State(port, cur);

	for (PathTransition* p = std_transitions; p != END(std_transitions); ++p)
		{
		if ((p->from == Token(cur)) && (p->to == tok))
			{
			tm.Run_TAP_X(port, p->bits, p->TMS);
			tm.Flush_Commands();
			return;
			}
		}

	error("not a stable state transition");
}

void IMPLEMENT::do_state(Token tok)
{
	tok = NextToken();

	if (PeekToken() == kTerm)
		{
		to_stable_state(tok);
		check_term();
		}
	else
		{
		while (tok <= kIRUPDATE)
			{
			tm.Do_State_Transition(port, Pod::TAP_State(tok));
			tok = NextToken();
			}

		// check final state
		Pod::TAP_State cur;
		tm.Get_Current_State(port, cur);
		switch (cur)
			{
		case Pod::IRPAUSE:
		case Pod::DRPAUSE:
		case Pod::IDLE:
		case Pod::RESET:
			break;

		default:
			error(string("did not end at a stable state"));
			}
		}
}

void IMPLEMENT::parse_bits(unsigned max_bits, BitString& b)
{
//	locale			loc;
//	const ctype<SimpleLexer::Char_t>& ct
//		(use_facet< ctype<SimpleLexer::Char_t> >(loc));

	b.data.clear();
	b.data.reserve((max_bits + b.MAX_SHIFT - 1) / b.MAX_SHIFT);

	if (NextToken() != kStartHex)
		error("expected '('");

	unsigned shift = b.MAX_SHIFT;
	BitString::Base_t cur = 0;
	unsigned ping;
	for (bool done = false; !done;)
		{
		SimpleLexer::Char_t c = lex->NextChar();
//		if (ct.is(ctype_base::space, c))
//			{ /* ignore */ }
//		else
			{
			switch (c)
				{
			case 'a': case 'b': case 'c': case 'd':
			case 'e': case 'f': case 'A': case 'B':
			case 'C': case 'D': case 'E': case 'F':
				c -= ('A' - '0' - 10);
				// fall-thru
			case '0': case '1': case '2': case '3':
			case '4': case '5': case '6': case '7':
			case '8': case '9':
				if (shift == 0)
					{
					b.data.push_back(cur);
					shift = b.MAX_SHIFT;
					cur = 0;
					}
				shift -= 4;
				cur |= (c & 0xF) << shift;
				break;

			case ')':
				done = true;				/* end of string */
				break;

			case ' ': case '\t': case '\r': case '\n':
				break;

			default:
				error("bad hex string");
				}
			}
		}

	b.shift = shift;
	if (shift < b.MAX_SHIFT)
		b.data.push_back(cur);

	b.present = true;
}

void IMPLEMENT::parse_register(RegisterData& r)
{
	// reset this flag - does not get carried over like masks or TDI
	r.TDO.present = false;

	if (Next() != SimpleLexer::rNumber)
		error("expected <length>");

	if (unsigned(lex->intVal) != r.bits)
		{ r.reset(); }

	r.bits = lex->intVal;
	for (;;)
		switch (NextToken())
			{
		case kTDI:
			parse_bits(r.bits, r.TDI);
			break;

		case kTDO:
			parse_bits(r.bits, r.TDO);
			break;

		case kMASK:
			parse_bits(r.bits, r.MASK);
			break;

		case kSMASK:
			// the purpose of SMASK is unclear to me... what is an 'invalid
			// output bit'?  Should they be forced to 0?
			parse_bits(r.bits, r.SMASK);
			break;

		case kTerm:
			return;

		default:
			error("Bad register syntax");
			}
}

void IMPLEMENT::do_header_trailer(Token tok)
{
	if (chain)
		{
		// ignore (parse into throwaway)
		parse_register(RegisterData());
		}
	else
		{
		switch (tok)
			{
		case kHIR: parse_register(reg[IDX_HIR]); break;
		case kHDR: parse_register(reg[IDX_HDR]); break;
		case kTIR: parse_register(reg[IDX_TIR]); break;
		case kTDR: parse_register(reg[IDX_TDR]); break;
			}
		}
}

void IMPLEMENT::do_scan(Token tok)
{
	int base_r = (tok == kSIR) ? IDX_HIR : IDX_HDR;

	unsigned start_text = lex->tellg();
	parse_register(reg[base_r + 1]);
	unsigned text_length = lex->tellg() - start_text;

	if (chain)
		{
		static bool valid_SIR = false;

		// if in a chain, inhibit SIRs that don't address just the
		// main device
		if (tok == kSIR)
			{
			if (!(valid_SIR = (chain->IR_bits == reg[base_r + 1].bits)))
				return;
			}
		// inhibit SDR's not preceeded by valid SIR's
		else if (!valid_SIR)
			return;
		}

	// move to appropriate SELECTx state
	Pod::TAP_State cur, target =
		(tok == kSIR) ? Pod::IRSELECT : Pod::DRSELECT;
	for (;;)
		{
		tm.Get_Current_State(port, cur);
		if (cur == target) break;

		// for most states, shifting a 1 will move through the interesting
		// states, but we need to shift a 0 to get out of RESET and
		// a 0 to move from SELECTIR to CAPTUREIR
		switch (cur)
			{
		case kIRSELECT:
		case kRESET:
			tm.Do_State_Move(port, 0);
			break;

		default:
			tm.Do_State_Move(port, 1);
			break;
			}
		}

	// CAPTUREx SHIFTx
	tm.Do_State_Move(port, 0);
	tm.Do_State_Move(port, 0);

	// Here is where all the work happens...
	// need to compose together the headers, data and trailers,
	// and shift the whole bit stream out

	unsigned buffer_bytes = caps.buffer_bits / 8;

	// allocate buffer if we need to capture TDO and verify it
	bool do_check_TDO = check_TDO && reg[base_r + IDX_IR].TDO.present;
	auto_ptr<BitComposer::Byte> inTDO
		(do_check_TDO ? new BitComposer::Byte[buffer_bytes] : 0);
	BitComposer::Byte* pTDO = inTDO.get();

	// allocate a streaming object for each component of the shift
	BitComposer	TDI(buffer_bytes), SMASK(buffer_bytes);
	BitComposer	TDO(buffer_bytes), MASK(buffer_bytes);

	unsigned bit_base, total_bits;
	int stream;
	for (total_bits = 0, stream = IDX_HIR; stream <= IDX_TIR; ++stream)
		total_bits += reg[base_r + stream].bits;

	unsigned text_incr;
	if (total_bits < caps.buffer_bits)
		text_incr = text_length;
	else
		text_incr = unsigned(text_length *
							(float(caps.buffer_bits) / float(total_bits)));

	bit_base = 0;
	for (stream = IDX_HIR; stream <= IDX_TIR;)
		{
		RegisterData& curReg = reg[base_r + stream];
		unsigned bits = curReg.bits;

		bool end_of_stream =
		TDI.Compose(bits, false, &curReg.TDI);
		SMASK.Compose(bits, curReg.TDI.present && !curReg.SMASK.present, &curReg.SMASK);
		TDO.Compose(bits, false, &curReg.TDO);
		MASK.Compose(bits, curReg.TDO.present && !curReg.MASK.present, &curReg.MASK);

		bool end_of_reg = end_of_stream && (stream == IDX_TIR);

		if (end_of_reg || (TDI.outp == TDI.limit))
			{
			if (end_of_reg)
				{
				// n.b. there is a reserved byte in each buffer
				// to hold this, if present
				TDI.Flush();
				TDO.Flush();
				MASK.Flush();
				SMASK.Flush();
				}

			// have some (or zero) output bits in the streaming objects,
			// send them out and optionally make sure what comes back is right too

			unsigned p;

			// first, combine TDI and SMASK
			for (p = TDI.outp - TDI.buffer.begin(); p--;)
				{ TDI.buffer[p] &= SMASK.buffer[p]; }

			// then run the TAP
			bits = min(total_bits, 8U * (TDI.outp - TDI.buffer.begin()));

			report_status(start_text += text_incr);

			total_bits -= bits;
			tm.Do_Shift(port, bits, end_of_reg, TDI.buffer.begin(), pTDO);

			// compare what was shifted in
			if (do_check_TDO)
				{
				// compare with TDO and MASK
				unsigned end_byte = TDO.outp - TDO.buffer.begin();
				for (p = 0; p < end_byte; ++p)
					if ((pTDO[p] ^ TDO.buffer[p]) & MASK.buffer[p])
						{
						ostrstream msg;
						msg << "TDO data mismatch during scan at bit position "
							<< bit_base + p*8
							<< "; expected 0x"
							<< hex << setfill('0') << setw(2)
							<< unsigned(TDO.buffer[p] & MASK.buffer[p])
							<< "; got 0x"
							<< hex << setfill('0') << setw(2)
							<< unsigned(pTDO[p] & MASK.buffer[p]);
						error(msg.str());
						}
				}

			bit_base += bits;

			// clear buffers for another round
			TDI.Clear_Buffer();
			TDO.Clear_Buffer();
			MASK.Clear_Buffer();
			SMASK.Clear_Buffer();
			}

		// move to next component of stream if necessary
		if (end_of_stream)
			{ ++stream; }
		}

	// move into requested final state
	unsigned dist = 0, moves;
	switch (reg[base_r + IDX_IR].end_state)
		{
	case kRESET:
		// UPDATEx SELECTDR SELECTIR RESET
		dist = 4;
		moves = 0xF;
		break;

	case kIDLE:
		// UPDATEx IDLE
		dist = 2;
		moves = 0x01;
		break;

	case kDRPAUSE:
		if (tok == kSIR)
			{	// from SHIFTIR
			// UPDATEIR SELECTDR CAPTUREDR EXIT1DR PAUSEDR
			dist = 5;
			moves = 0x0B;
			}
		else
			{	// from SHIFTDR
			// PAUSEDR
			dist = 1;
			moves = 0x00;
			}
		break;

	case kIRPAUSE:
		if (tok == kSIR)
			{	// from SHIFTIR
			// PAUSEIR
			dist = 1;
			moves = 0x00;
			}
		else
			{	// from SHIFTDR
			// UPDATEDR SELECTDR SELECTIR CAPTUREIR EXIT1IR PAUSEIR
			dist = 6;
			moves = 0x17;
			}
		break;
		}

	tm.Run_TAP_X(port, dist, moves);
	tm.Flush_Commands();
}

void IMPLEMENT::parse_real(double& result)
{
	int int_part, frac_part = 0, exp_part = 0;

	if (Next() != SimpleLexer::rNumber)
		error("expected number");

	int_part = lex->intVal;

	if (PeekToken() == kDot)
		{
		NextToken();
		if (Peek() != SimpleLexer::rNumber)
			error("expected digits after decimal");
		frac_part = lex->intVal;
		NextToken();
		}

	if ((lex->Peek() == SimpleLexer::rIdent) &&
		((lex->stringVal[0] == 'e') || (lex->stringVal[0] == 'E')))
		{
		static string EN("expected exponent after 'E'");

		bool exp_minus = false;
		string::size_type pos = 1;

		if (lex->stringVal.length() == 1)
			{	// if we only have 'E', must be either -exponent or error
			Next(); 	// skip 'E'
			if ((NextToken() != kMinus) ||
				(Peek() != SimpleLexer::rNumber))
				error(EN);

			exp_minus = true;
			pos = 0;
			}

		exp_part = 0;
		while (pos < lex->stringVal.length())
			{
			string::value_type c = lex->stringVal[pos++];
			if ((c < '0') || (c > '9'))
				error(EN);
			exp_part = (exp_part * 10) + (c - '0');
			}

		if (exp_minus)
			exp_part = -exp_part;
		NextToken();
		}

	int frac_dig = frac_part ? int(ceil(log(frac_part) / log(10))) : 0;

	result = (double(int_part) + frac_part / pow(10, frac_dig)) * pow(10, exp_part);
}

void IMPLEMENT::do_frequency(Token tok)
{
	Pod::TAP_Settings cur_ts;
	tm.Get_TAP_Settings(port, cur_ts);

	if (PeekToken() == kTerm)
		{
		// return to default clock
		cur_frequency = cur_ts.TCK_rate_index = initial_frequency;
		tm.Set_TAP_Settings(port, cur_ts);
		}
	else
		{
		double freq;
		parse_real(freq);
		if (NextToken() != kHZ)
			error("Expected HZ");

		// find the fast TCK clock below the requested frequency
		unsigned index = 0;
		while ((index < caps.num_TCK_rates) &&
			   (freq < caps.TCK_rates[index]))
			{ ++index; }

		if (index == caps.num_TCK_rates)
			error("requested FREQUENCY too low - can't do it!");

		cur_frequency = cur_ts.TCK_rate_index = index;
		tm.Set_TAP_Settings(port, cur_ts);
		}
	check_term();
}

void IMPLEMENT::do_runtest(Token tok)
{
	double cycles, min_sec = 0.0, max_sec = 0.0;

	// parse [run_state]?
	if (PeekToken() <= kIRUPDATE)
		{
		runtest_run_state = NextToken();
		if (runtest_run_state == kRESET)
			error("RUNTEST RESET not supported");
		}

	// parse [run_count run_clk] | [min_time SEC]
	parse_real(cycles);

	switch (PeekToken())
		{
	case kTCK:
		NextToken();
		break;

	case kSEC:
		NextToken();
		min_sec = cycles;
		cycles = 0.0;
		break;

	case kSCK:
		error("System clock SCK not supported");

	default:
		error("expected TCK, SCK, or SEC");
		};

	// parse [min_time SEC]?
	if (Peek() == SimpleLexer::rNumber)
		{
		// must be xx SEC
		parse_real(min_sec);
		if (NextToken() != kSEC)
			error("expected SEC");
		}

	// parse [MAXIMUM max_time SEC]?
	if (PeekToken() == kMAXIMUM)
		{
		NextToken();
		parse_real(max_sec);
		if (NextToken() != kSEC)
			error("expected SEC");
		}

	// parse [ENDSTATE state]?
	if (PeekToken() == kENDSTATE)
		{
		NextToken();
		if (PeekToken() > kIRUPDATE)
			error("expected state name");
		runtest_end_state = NextToken();
		}

	// and ending with a terminator
	check_term();

	// shift into run_state if necessary
	Pod::TAP_State cur;
	tm.Get_Current_State(port, cur);
	if (cur != runtest_run_state)
		to_stable_state(runtest_run_state);

	// calculate actual number of TCKs we will do

	// how long is one TCK?
	double period =
		1.0 / (Xilinx_timing ? XILINX_DEFAULT_TCK : caps.TCK_rates[cur_frequency]);

	// increase count if less than min_sec
	if (cycles * period < min_sec)
		cycles = min_sec / period;

	// decrease count if max_sec specified and more than it
	if ((max_sec > min_sec) && (cycles * period > max_sec))
		cycles = max_sec / period;

	Pod::TAP_Settings info;

	// is this a 'big run'?
	if (cycles > caps.buffer_bits)
		{
		// yes; use the timed, approximate TCK count method
		tm.Do_Run_TCK(port, true);
		Sleep(unsigned(cycles * period * 1000));
		tm.Do_Run_TCK(port, false);
		}
	else
		{
		// no, do an exact number of TCK cycles
		tm.Do_Shift(port, unsigned(cycles), false, NULL, NULL);
		tm.Flush_Commands();
		}

	// do transition to end state if necessary
	if (runtest_end_state != runtest_run_state)
		to_stable_state(runtest_end_state);
}

}

