//---------------------------------------------------------------------------

#if defined(__WIN32__)
  #include <windows.h>
#endif

#include "SimpleLexer.h"
#include <map>
#include <locale>

using namespace std;

namespace JTAG
{

static char alnum[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

SimpleLexer::SetOfChar SimpleLexer::UCALPHA(alnum, alnum + 26);
SimpleLexer::SetOfChar SimpleLexer::NUMBERS(alnum + 26, alnum + 36);
SimpleLexer::SetOfChar SimpleLexer::UCALPHANUM(alnum, alnum + 36);

class SimpleLexer::Data
	{
public:

	enum { MAX_BUFFER = 1U<<18 };

	Data
		(const Keyword*		a_keywords_begin,
		 const Keyword*		a_keywords_end);

	class LexRec
		{
	public:
		LexerResult	in_class;
		unsigned	index;

		LexRec() : in_class(rNone), index(0) {}

		bool operator < (const LexRec& b) const
			{ if (in_class < b.in_class) return true;
			  return (index < b.index); };
		};


	struct LUCmp { bool operator () (const Keyword* a, const Keyword* b)
		const { return *a < *b; } };
	typedef map<const Keyword*, LexRec, LUCmp> Lookup;

	Lookup			lu;
	char_traits<Char_t>::int_type	next;
	locale			loc;
	const ctype<Char_t>&	ct;

	Char_t	buffer[MAX_BUFFER], *cur, *end;
	bool			last_buffer;
	};

SimpleLexer::Data::Data
	(const Keyword*		a_keywords_begin,
	 const Keyword*		a_keywords_end)
	:	next(0),
		ct(use_facet< ctype<Char_t> >(loc))
{
	LexRec val;

	val.in_class = rKeyword;
	for (val.index = 0;
		 a_keywords_begin != a_keywords_end;
		 ++a_keywords_begin, ++val.index)
		lu.insert(Lookup::value_type(a_keywords_begin, val));

	val.in_class = rEOL;
	static string CR("\r"), LF("\n");
	lu.insert(Lookup::value_type(&CR, val));
	lu.insert(Lookup::value_type(&LF, val));

	val.in_class = rWhite;
	static string TAB("\t"), SPACE(" ");
	lu.insert(Lookup::value_type(&TAB, val));
	lu.insert(Lookup::value_type(&SPACE, val));
}

unsigned SimpleLexer::tellg()
{
	return already_read + (cur - data->buffer);
}

void SimpleLexer::check_buffer()
{
	if (!last_buffer)
		{
		if (cur)
			already_read += (cur - data->buffer);
		cur = data->buffer;
		input->read(cur, data->MAX_BUFFER);
		end = data->buffer + input->gcount();
		last_buffer = (end != &data->buffer[data->MAX_BUFFER]);
		}
}

SimpleLexer::SimpleLexer
	(const Keyword*		a_keywords_begin,
	 const Keyword*		a_keywords_end,
	 bool			a_case_sensitive,
	 const SetOfChar&	a_ident_start,
	 const SetOfChar&	a_ident_rest)
	: 	last(rNone),
		case_sensitive(a_case_sensitive),
		ident_start(a_ident_start),
		ident_rest(a_ident_rest),
		data (new Data(a_keywords_begin, a_keywords_end)),
		last_buffer(false), cur(0), end(0),
		line(1), column(1)
{}

void SimpleLexer::Open(std::istream&	a_input)
{
	input = &a_input;
	check_buffer();
}

int SimpleLexer::peek2()
{
	check_buffer();
	if (cur == end)
		return EOF;
	return *cur;
}

inline int SimpleLexer::PeekChar()
{
	if (cur == end) return peek2();
	return *cur;
}

inline int SimpleLexer::NextChar()
{
	int c = PeekChar();
	if (c != EOF) ++cur;
	column = (c == '\n') ? (++line, 0) : (column + 1);
	return c;
}

inline void SimpleLexer::EatPeekChar()
{
	(void) NextChar();
}

void SimpleLexer::SkipWS()
{
	Keyword s;
	do
		{
		int c = PeekChar();
		if (c == EOF) return;
		s = char_traits<Char_t>::to_char_type(c);
		Data::Lookup::iterator p = data->lu.find(&s);
		if ((p == data->lu.end()) || (p->second.in_class != rWhite))
			break;
		EatPeekChar();
		}
	while (true);
}

SimpleLexer::LexerResult SimpleLexer::Lex1()
{
	SkipWS();
	int c0 = PeekChar();
	if (c0 == EOF) return rEOF;

	Char_t c = char_traits<Char_t>::to_char_type(c0);
	if (!case_sensitive)
		c = data->ct.toupper(c);
	stringVal = c;

	if (data->ct.is(ctype_base::digit, c))
		{
		// a number
		intVal = c - Char_t('0');
		for (;;)
			{
			EatPeekChar();
			if (EOF == (c0 = PeekChar())) break;
			c = char_traits<Char_t>::to_char_type(c0);
			if (!data->ct.is(ctype_base::digit, c)) break;
			stringVal += c;
			intVal = (intVal * 10) + (c - Char_t('0'));
			}
		return rNumber;
		}
	else
		{
		if (ident_start.count(c) > 0)
			{
			// an identifier or keyword
			for (;;)
				{
				EatPeekChar();
				if (EOF == (c0 = PeekChar())) break;
				c = char_traits<Char_t>::to_char_type(c0);
				if (!case_sensitive)
					c = data->ct.toupper(c);
				if (ident_rest.count(c) == 0) break;
				stringVal += c;
				}

			Data::Lookup::iterator p = data->lu.find(&stringVal);
			if (p != data->lu.end())
				{
				index = p->second.index;
				return p->second.in_class;
				}
			return rIdent;
			}
		else
			{
			// something else!
			Data::Lookup::iterator pos;
			for (unsigned len = 1; ; ++len)
				{
				pos = data->lu.lower_bound(&stringVal);
				if ((pos != data->lu.end()) &&
					(stringVal[len - 1] == (*pos->first)[len - 1]))
					{
					EatPeekChar();
					if (len == pos->first->length())
						{
						index = pos->second.index;
						return pos->second.in_class;
						}
					if (EOF == (c0 = PeekChar())) break;
					c = char_traits<Char_t>::to_char_type(c0);
					stringVal += c;
					}
				else
					break;
				}
			}
		}

	EatPeekChar();
	return rUnknown;
}

SimpleLexer::LexerResult SimpleLexer::Peek()
{
	if (last == rNone)
		{ last = Lex1(); }
	return last;
}

SimpleLexer::LexerResult SimpleLexer::Next()
{
	LexerResult res = Peek();
	last = rNone;
	return res;
}

}

#if defined(__WIN32__)
#pragma argsused
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fwdreason, LPVOID lpvReserved)
{
	return 1;
}
#endif

//---------------------------------------------------------------------------

