//---------------------------------------------------------------------------

#include <vector>
#include <memory>
#include <strstream>
#include <iomanip>
#include <assert>

#if defined(__WIN32__)
  #define WIN32_LEAN_AND_MEAN
  #include <windows.h>
  #include "FTD2XX.H"
  #pragma link "FTD2XX.LIB"
#endif

#pragma hdrstop

//#define DEBUG

#include "Pod.h"

// This interface class uses the FTD2XX library from FTDI.  Under other
// operating systems than Windows, it may be necessary to implement a
// subset of the functions that it defines.

using namespace std;

#define ELEMS(x) sizeof(x)/sizeof(x[0])

namespace JTAG
{

const double CLOCK_PERIOD = 62.5e-9;

enum Commands
	{
	CMD_SETCURTAP = 0x00,
	CMD_GETSTATUS = 0x10,
	CMD_WRBUF = 0x20,
	CMD_RDBUF = 0x30,
	CMD_RUN = 0x40,
		BIT_TDI = 0x01,
		BIT_TDO = 0x02,
		BIT_TMS = 0x04,
	CMD_COPY = 0x50,
	CMD_DELAY = 0x60,
	REG_GLOBAL = 0x80,
	REG_COUNT = 0x90,
	REG_TMS = 0xA0,
	REG_BITS = 0xB0,
	REG_ADDR = 0xC0,
	REG_CONFIG1 = 0xE0,
	REG_CONFIG2 = 0xF0,
		BIT_TRISTATE = 0x80,
		BIT_RST = 0x100,
		BIT_TRST = 0x200,
		BIT_DINT = 0x400,
		BIT_RUNTCK = 0x800
	};

struct HW_Info
	{
	char		magic[4];
	unsigned    HW_rev;
	USB_Pod_FPGA_Types	FPGA_type;
	};

#define MAGIC "UJP1"

class USBPod_I
	: public Pod
	{
public:

// Interface

#define TAPMGRIF(s) void s;

	TAPMGRIF( Get_General_Info(Device_Capabilities&) )
	TAPMGRIF( Set_LED(unsigned, bool) )

	TAPMGRIF( Flush_Commands() )

	TAPMGRIF( Get_TAP_Info(TAPPort, TAP_Capabilities&) )
	TAPMGRIF( Get_TAP_Settings(TAPPort, TAP_Settings&) )
	TAPMGRIF( Set_TAP_Settings(TAPPort, TAP_Settings&) )

	TAPMGRIF( Get_Current_State(TAPPort, TAP_State&) )
	TAPMGRIF( Do_State_Transition(TAPPort, TAP_State = RESET) )
	TAPMGRIF( Do_State_Move(TAPPort, int = 1 /* 0 or 1 */) )
	TAPMGRIF( Do_Shift(TAPPort, unsigned bits, bool end_of_shift,
					   Byte* TDO, Byte* TDI = 0) )
	TAPMGRIF( Do_Run_TCK(TAPPort, bool onoff) )

	TAPMGRIF( Write_TDO_Buffer(TAPPort, unsigned offset, unsigned bytes, Byte*) )
	TAPMGRIF( Read_TDI_Buffer(TAPPort, unsigned offset, unsigned bytes, Byte*) )
	TAPMGRIF( Copy_TDI_Buffer_To_TDO(TAPPort, unsigned offset, unsigned bytes) )
	TAPMGRIF( Run_TAP_X(TAPPort, unsigned bits, unsigned TMS, unsigned flags = 0) )

	TAPMGRIF( Set_TRST(TAPPort, bool) )
	TAPMGRIF( Pulse_TRST(TAPPort, unsigned ms) )

	TAPMGRIF( Set_RST(TAPPort, bool) )
	TAPMGRIF( Pulse_RST(TAPPort, unsigned ms) )

	TAPMGRIF( Set_DINT(TAPPort, bool) )
	TAPMGRIF( Pulse_DINT(TAPPort, unsigned ms) )

	TAPMGRIF( Get_TAP_GPIO_Input(TAPPort, unsigned, GPIO&) )
	TAPMGRIF( Set_TAP_GPIO(TAPPort, unsigned, GPIO) )
	TAPMGRIF( Pulse_TAP_GPIO(TAPPort, unsigned, unsigned ns, GPIO pulse, GPIO after) )

// Implementation

	void Reset();

	USBPod_I(unsigned dev);
	~USBPod_I();

	void Read(Byte*, unsigned);
	void Write(Byte*, unsigned);
	void Write_Imm(Byte*, unsigned);

	void UpdateGLOBAL();

	enum
		{
		CONFIGBUFSIZE = 32768,
		TMS_MAX_BITS = 12,
		MAX_XFER = 550
		};

	static TAP_State state_transition[16][2];

	HW_Info		HW;
	FT_HANDLE	driver;
	char		ID_string[32];

	Byte		out_buffer[MAX_XFER];
	Byte*		out_pos;

	unsigned	GLOBAL;

	struct CurrentTAP
		{
		GPIO		GPIOs[4];
		TAP_Settings	TAP;
		unsigned	CONFIG1;
		unsigned	CONFIG2;

		TAP_State	state;
#if defined(DEBUG)
		TAP_State	last_state;
#endif
		unsigned	TMS;
		unsigned	TMS_bits;

		void UpdateCONFIG();
		}		current[2];

	void Flush_TMS(CurrentTAP*);
	void UpdateTAP(CurrentTAP*);
	void Pulse(CurrentTAP*, unsigned X, unsigned ms);
	void Set(CurrentTAP*, unsigned X, bool on);
	};

#include "fpga_30.c"
#include "fpga_50.c"
#include "fpga_100.c"

static unsigned char *bitfiles[] = { bitfile30, bitfile50, bitfile100 };

USBPod_I::TAP_State USBPod_I::state_transition[16][2] =
	{
	/*RESET*/		{ IDLE, RESET },
	/*IDLE*/		{ IDLE, DRSELECT },
	/*DRSELECT*/	{ DRCAPTURE, IRSELECT },
	/*DRCAPTURE*/	{ DRSHIFT, DREXIT1 },
	/*DRSHIFT*/		{ DRSHIFT, DREXIT1 },
	/*DREXIT1*/		{ DRPAUSE, DRUPDATE },
	/*DRPAUSE*/		{ DRPAUSE, DREXIT2 },
	/*DREXIT2*/		{ DRSHIFT, DRUPDATE },
	/*DRUPDATE*/	{ IDLE, DRSELECT },
	/*IRSELECT*/	{ IRCAPTURE, RESET },
	/*IRCAPTURE*/	{ IRSHIFT, IREXIT1 },
	/*IRSHIFT*/		{ IRSHIFT, IREXIT1 },
	/*IREXIT1*/		{ IRPAUSE, IRUPDATE },
	/*IRPAUSE*/		{ IRPAUSE, IREXIT2 },
	/*IREXIT2*/		{ IRSHIFT, IRUPDATE },
	/*IRUPDATE*/	{ IDLE, DRSELECT }
	};

static vector<DWORD>	units;

//---------------------------------------------------------------------------

void CountDevices
	(unsigned& res)
{
	units.clear();
	for (DWORD idx = 0; idx < 99; ++idx)
		{
		char buffer[64];
		switch (FT_ListDevices(PVOID(idx), &buffer, FT_LIST_BY_INDEX|FT_OPEN_BY_DESCRIPTION))
			{
		case FT_OK:
			if (strncmp(buffer, "USB JTAG", 8) == 0)
				units.push_back(idx);
			break;

		case FT_DEVICE_NOT_FOUND:
			idx = 999;
			break;

		default:
			throw Error_Exception("USB driver not loaded");
			}
		}

	res = units.size();
}

Pod* new_Pod
	(unsigned index)
{
	return new USBPod_I(index);
}

Pod::~Pod
	()
{}

static void check(FT_STATUS code)
{
	if (code != FT_OK)
		{
		ostrstream msg;
		msg << "USB Driver error (code " << code << ")";
		throw Error_Exception(msg.str());
		}
}

void ConfigureHW
	(unsigned device, unsigned HW_rev, USB_Pod_FPGA_Types type)
{
	FT_HANDLE driver = 0;

	unsigned num;
	if (units.empty())
		CountDevices(num);

	if (device >= num)
		throw Error_Exception("bad device #");

	// attempt to open and program USB JTAG device
	check(FT_Open(units[device], &driver));

	try
		{
		DWORD wrote;
		check(FT_SetTimeouts(driver, 500, 500));
		check(FT_SetLatencyTimer(driver, 1));

		// debug
		//FT_EE_UASize(driver, &wrote);

		HW_Info HW = { MAGIC, HW_rev, type };

		// retrieve and check EEUA data area
		check(FT_EE_UAWrite(driver, (PUCHAR)&HW, sizeof(HW)));
		}

	catch (...)
		{
		FT_Close(driver);
		throw;
		}

	FT_Close(driver);
}

//---------------------------------------------------------------------------

USBPod_I::USBPod_I(unsigned dev)
	: driver(0), out_pos(out_buffer)
{
	unsigned num;
	if (units.empty())
		CountDevices(num);

	if (dev >= units.size())
		throw Error_Exception("bad device #");

	// attempt to open and program USB JTAG device
	check(FT_Open(units[dev], &driver));

	try
		{
		DWORD wrote;
		check(FT_SetTimeouts(driver, 5000, 5000));

		// use a bigger transmit size for configuration
		check(FT_SetUSBParameters(driver, 65536, 65536));

		// retrieve and check EEUA data area
		check(FT_EE_UARead(driver, (PUCHAR)&HW, sizeof(HW), &wrote));

		if ((0 != memcmp(HW.magic, MAGIC, sizeof(MAGIC))) ||
			(HW.FPGA_type > ELEMS(bitfiles)))
			throw Error_Exception("HW config not set");

		// first attempt to enter reconfiguration using the
		// command code, in case the FPGA is already configured
		check(FT_Purge(driver, FT_PURGE_RX | FT_PURGE_TX));
		//GLOBAL = 0x900;
		//UpdateGLOBAL();
		//Sleep(4);

		// get appropriate bitfile for FPGA & PCB rev
		// (currently same bitfile for different HW revs)
		unsigned char* pbitfile = bitfiles[HW.FPGA_type];

		// skip over bitfile headers
		for (int skip1 = 3; skip1--;)
			pbitfile += 2 + (pbitfile[0] << 8) + pbitfile[1];

		for (;;)
			{
			if (pbitfile[0] == 'e') break;
			pbitfile += 3 + (pbitfile[1] << 8) + pbitfile[2];
			}

		// get length
		unsigned bitlength = (unsigned(pbitfile[1]) << 24) |
							 (unsigned(pbitfile[2]) << 16) |
							 (unsigned(pbitfile[3]) <<  8) |
							  unsigned(pbitfile[4]);
		pbitfile += 5;

		// enter bit-bang mode, top speed
		check(FT_SetBaudRate(driver, FT_BAUD_921600));
		check(FT_SetBitMode(driver, 0xFF, 1));

		// convert to bit bangs and send out
		vector<char> buf;
		buf.resize(CONFIGBUFSIZE);
		char *p = buf.begin();
		while (bitlength--)
			{
			char c = *pbitfile++;
			for (int n = 8; n--;)
				{
				p[1] = (p[0] = ((c >> 7) & 1)) | 0x80;
				p += 2;
				c <<= 1;
				}

			if ((p == buf.end()) || (bitlength == 0))
				{
				Write_Imm(buf.begin(), p - buf.begin());
				p = buf.begin();
				}
			}

		// exit bit-bang mode
		check(FT_SetBitMode(driver, 0xFF, 0));

		// biggest chunk we'll transmit is 511 bytes plus some commands
		check(FT_SetUSBParameters(driver, MAX_XFER, MAX_XFER));

		// wait for open_status to be returned and freq. lock to be true
		int retry = 20;
		for (;;)
			{
			char getstatus[] = { CMD_GETSTATUS };

			check(FT_Purge(driver, FT_PURGE_RX));
			Write_Imm(getstatus, sizeof(getstatus));

			char res;
			Read(&res, 1);
			if (res & 0x02)
				break;

			if (--retry == 0)
				throw Error_Exception("FPGA didn't start up");
			}

		// Retrieve FPGA-ware version string
		// This is stored in the TDI buffer memory
		// at initialization, so we just have to read it

		char readID[] = { REG_COUNT, sizeof(ID_string),
							REG_ADDR, 0,
							CMD_RDBUF };

		Write_Imm(readID, sizeof(readID));
		Read(ID_string, sizeof(ID_string));

		// init TAPs on pod
		Reset();
		}

	catch (...)
		{
		FT_Close(driver);
		throw;
		}
}

USBPod_I::~USBPod_I
	()
{
	if (driver)
		{
		/// de-configure
		GLOBAL = 0x900;
		UpdateGLOBAL();
		// and close
		Flush_Commands();
		FT_Close(driver);
		}
}

void USBPod_I::Read
	(Byte* mem, unsigned len)
{
	Flush_Commands();
	DWORD read = 0;
	check(FT_Read(driver, mem, len, &read));
	if (read != len)
		throw Error_Exception("FT_Read() short");
}

void USBPod_I::Write_Imm
	(Byte* mem, unsigned len)
{
	DWORD wrote = 0;
	check(FT_Write(driver, mem, len, &wrote));
	if (wrote != len)
		throw Error_Exception("FT_Write() short");
}

void USBPod_I::Flush_Commands
	()
{
	if (out_pos > out_buffer)
		{
		//!!! unless FT_Write is guaranteed to be atomic,
		//!!! needs to be a critical area here

		DWORD len = out_pos - out_buffer;
		out_pos = out_buffer;
		Write_Imm(out_buffer, len);
		}
}

void USBPod_I::Write
	(Byte* mem, unsigned len)
{
	if (out_pos + len > &out_buffer[MAX_XFER])
		Flush_Commands();

	assert(len <= MAX_XFER);

	memcpy(out_pos, mem, len);
	out_pos += len;
}

void USBPod_I::UpdateGLOBAL
	()
{
	char updateglobal[] = { REG_GLOBAL | ((GLOBAL & 0xF00)>>8), GLOBAL & 0xFF };
	Write(updateglobal, sizeof(updateglobal));
}

void USBPod_I::CurrentTAP::UpdateCONFIG
	()
{
	CONFIG1 = (REG_CONFIG1 << 8) |
				 (TAP.TCK_term << 10) |
				 (TAP.TDO_term << 9) |
				 (GPIOs[0]) |
				 (GPIOs[1] << 2) |
				 (GPIOs[2] << 4) |
				 (GPIOs[3] << 6);

	CONFIG2 = (REG_CONFIG2 << 8) |
				 BIT_RST |
				 BIT_TRST |
				 (TAP.tristated ? BIT_TRISTATE : 0) |
				 (TAP.TCK_rate_index) |
				 (TAP.current_TDO_buffer << 4);
}

void USBPod_I::UpdateTAP
	(CurrentTAP* now)
{
	now->UpdateCONFIG();

	char updateTAP[] =
		{
		CMD_SETCURTAP | (now - current),
		char(now->CONFIG1 >> 8), char(now->CONFIG1),
		char(now->CONFIG2 >> 8), char(now->CONFIG2)
		};

	Write(updateTAP, sizeof(updateTAP));
}

void USBPod_I::Get_General_Info
	(Device_Capabilities& info)
{
	// for now, we have only one implementation,
	// so these are (mostly) fixed values
	static const float TCKs[] =
		{
		48e6, 32e6, 24e6, 16e6,
		12e6, 8e6, 6e6, 4e6,
		3e6, 2e6, 1.5e6, 1e6,
		750e3, 500e3, 375e3, 250e3
		};

	info.ID_string = ID_string;
	info.num_TAPs = 2;
	info.num_buffers = 4;
	info.buffer_bits = 4095;
	info.num_LEDs = 4;
	info.num_TCK_rates = 16;
	info.TCK_rates = TCKs;
	info.TMS_bits = TMS_MAX_BITS;
};

void USBPod_I::Get_TAP_Info
	(TAPPort port, TAP_Capabilities& info)
{
	info.num_GPIOs = 4;
	info.canTristate = true;
	info.hasTCKTerm = (port == 0);
	info.hasTDOTerm = (port == 0) && (HW.HW_rev > 0);
}

void USBPod_I::Get_TAP_Settings
	(TAPPort port, TAP_Settings& info)
{
	info = current[port].TAP;

	char open_status[] = { CMD_SETCURTAP | port, CMD_GETSTATUS };

	//!!! needs to be a critical area here

	Write(open_status, sizeof(open_status));
	Read(open_status, 1);
	info.ready = (open_status[0] & 0x02) && !(open_status[0] & 0x01);
}

void USBPod_I::Set_TAP_Settings
	(TAPPort port, TAP_Settings& info)
{
	current[port].TAP = info;
	UpdateTAP(&current[port]);
}

void USBPod_I::Reset
	()
{
	TAPPort port;

	// set up default settings into shadow registers
	GLOBAL = 0;
	UpdateGLOBAL();
	for (port = 2; port--;)
		{
		for (int idx = 4; idx--;) current[port].GPIOs[idx] = isZ;
		current[port].TAP.tristated = false;
		current[port].TAP.current_TDO_buffer = port;
		current[port].TAP.TCK_term = TAP_Settings::term_33;
		current[port].TAP.TDO_term = TAP_Settings::unterminated;
		current[port].TAP.TCK_rate_index = 11;	// 1 MHz
		current[port].TAP.Current_GPIO_Setting = current[port].GPIOs;

		current[port].state = RESET;
		current[port].TMS = 0;
		current[port].TMS_bits = 0;

		UpdateTAP(&current[port]);
		}
}

void USBPod_I::Write_TDO_Buffer
	(TAPPort port, unsigned offset, unsigned bytes, Byte* TDO)
{
	char setTDO[] =
		{ CMD_SETCURTAP | port,
		  REG_ADDR | ((offset >> 8) & 0xF), offset,
		  REG_COUNT | ((bytes >> 8) & 0xF), bytes,
		  CMD_WRBUF };

	//!!! needs to be a critical area here

	Write(setTDO, sizeof(setTDO));
	Write(TDO, bytes);
}

void USBPod_I::Read_TDI_Buffer
	(TAPPort port, unsigned offset, unsigned bytes, Byte* TDI)
{
	//!!! needs to be a critical area here

	char getTDI[] =
		{ REG_ADDR | ((offset >> 8) & 0xF), offset,
		  REG_COUNT | ((bytes >> 8) & 0xF), bytes,
		  CMD_RDBUF };
	Write(getTDI, sizeof(getTDI));
	Read(TDI, bytes);
}

void USBPod_I::Flush_TMS
	(CurrentTAP* now)
{
#if defined(DEBUG)
	{
	ostrstream s;
	s << " <" << now->last_state << "> ";
	for (unsigned n = 0; n < now->TMS_bits; ++n)
		s << char('0' + ((now->TMS >> n)&1));
	s << " <" << now->state << ">";
	OutputDebugString(s.str());
	}
#endif

	char doRun[] =
		{ CMD_SETCURTAP | (now - current),
		  REG_TMS | ((now->TMS >> 8) & 0xF), now->TMS & 0xFF,
		  REG_BITS | ((now->TMS_bits >> 8) & 0xF), now->TMS_bits & 0xFF,
		  CMD_RUN };

	now->TMS = 0;
	now->TMS_bits = 0;

	Write(doRun, sizeof(doRun));
}

void USBPod_I::Run_TAP_X
	(TAPPort port, unsigned bits, unsigned TMS, unsigned flags)
{
	if (current[port].TMS_bits > 0)
		Flush_TMS(&current[port]);

	TAP_State& now = current[port].state;
#if defined(DEBUG)
	TAP_State prev_state = now;
#endif
	unsigned tempTMS = (TMS &= ((1<<TMS_MAX_BITS)-1));
	for (int n = bits; n--;)
		{
		if (((tempTMS == 0) || (tempTMS == ((1<<TMS_MAX_BITS)-1))) &&
			 (state_transition[now][tempTMS & 1] == now))
			break;	// optimization - in stable state
		now = state_transition[now][tempTMS & 1];
		tempTMS = (tempTMS & (1<<(TMS_MAX_BITS-1))) | (tempTMS >> 1);
		}

	if (flags & BIT_TMS)
		now = state_transition[now][1];

#if defined(DEBUG)
	{
	ostrstream s;
	s << " <" << prev_state << "> ";
	for (unsigned n = 0; n < min(bits, 12U); ++n)
		s << char('0' + ((TMS >> n)&1));
	if (bits > 12)
		s << " ... (" << bits << " bits)";
	s << " <" << now << ">";
	OutputDebugString(s.str());
	}
#endif

	char doRun[] =
		{ CMD_SETCURTAP | port,
		  REG_TMS | ((TMS >> 8) & 0xF), TMS & 0xFF,
		  REG_BITS | ((bits >> 8) & 0xF), bits & 0xFF,
		  CMD_RUN | flags };
	Write(doRun, sizeof(doRun));
}

void USBPod_I::Get_Current_State
	(TAPPort port, TAP_State& state)
{
	state = current[port].state;
}

void USBPod_I::Do_State_Transition
	(TAPPort port, TAP_State to_state)
{
	CurrentTAP& now = current[port];

	switch (now.TMS_bits)
		{
#if defined(DEBUG)
	case 0: now.last_state = now.state; break;
#endif
	case TMS_MAX_BITS: Flush_TMS(&now); break;
		}

	if ((to_state == state_transition[now.state][1]) || (to_state == RESET))
		now.TMS |= (1U << now.TMS_bits);
	else if (to_state != state_transition[now.state][0])
		throw Error_Exception("illegal transition");

	++now.TMS_bits;
	now.state = to_state;
}

void USBPod_I::Do_State_Move
	(TAPPort port, int dir /* 0 or 1 */)
{
	if ((dir < 0) || (dir > 1))
		throw Error_Exception("illegal transition");

	CurrentTAP& now = current[port];

	switch (now.TMS_bits)
		{
#if defined(DEBUG)
	case 0: now.last_state = now.state; break;
#endif
	case TMS_MAX_BITS: Flush_TMS(&now); break;
		}

	now.TMS |= dir << now.TMS_bits;
	++now.TMS_bits;
	now.state = state_transition[now.state][dir];
}

void USBPod_I::Do_Shift
	(TAPPort port, unsigned bits, bool end_of_shift, Byte* TDO, Byte* TDI)
{
	if (bits > 0)
		{
#if defined(DEBUG)
		if (TDO)
			{
			unsigned bytes = (bits+7)/8;
			ostrstream s;
			s << " -> ";
			if (bytes > 24U)
				{
				bytes = 24U;
				s << "... ";
				}
			while (bytes--)
				s << setw(2) << setfill('0') << hex
				  << unsigned(TDO[bytes]);
			OutputDebugString(s.str());
			}
#endif

		unsigned bytes = (bits + 7) / 8;

		//!!! needs to be a critical area here

		if (TDO)
			Write_TDO_Buffer(port, 0, bytes, TDO);

		unsigned flags = (TDI ? BIT_TDI : 0) |
						 (TDO ? BIT_TDO : 0) |
						 (end_of_shift ? BIT_TMS : 0);

		Run_TAP_X(port, bits, 0x000, flags);

		if (TDI)
			Read_TDI_Buffer(port, 0, bytes, TDI);

#if defined(DEBUG)
		if (TDI)
			{
			unsigned bytes = (bits+7)/8;
			ostrstream s;
			s << " <- ";
			if (bytes > 24U)
				{
				bytes = 24U;
				s << "... ";
				}
			while (bytes--)
				s << setw(2) << setfill('0') << hex
				  << unsigned(TDI[bytes]);
			OutputDebugString(s.str());
			}
#endif
		}
}

void USBPod_I::Set_LED(unsigned which, bool on)
{
	if (which > 3)
		throw Error_Exception("bad parameter");

	if (on)
		GLOBAL |= (1 << which);
	else
		GLOBAL &= ~(1 << which);

	UpdateGLOBAL();
	Flush_Commands();
}

void USBPod_I::Get_TAP_GPIO_Input
	(TAPPort port, unsigned n, GPIO& pin)
{
	pin = is0;

	char open_status[] = { CMD_SETCURTAP | port, CMD_GETSTATUS };

	//!!! needs to be a critical area here

	Write(open_status, sizeof(open_status));
	Read(open_status, 1);

	pin = (open_status[0] & (1<<(4+n))) ? is1 : is0;
}

void USBPod_I::Set_TAP_GPIO
	(TAPPort port, unsigned n, GPIO pin)
{
	if (n > 3)
		throw Error_Exception("bad parameter");

	current[port].GPIOs[n] = pin;
	UpdateTAP(&current[port]);
	Flush_Commands();
}

void USBPod_I::Pulse_TAP_GPIO
	(TAPPort port, unsigned n, unsigned ns, GPIO pulse, GPIO after)
{
	char seq[10], *out = seq;
	int delay;

	*out++ = CMD_SETCURTAP | port;

	if (ns < 800)
		delay = -1;
	else if (ns >= (1<<11)*125)
		delay = (1<<12)-1;
	else
		delay = int((ns - 800) / 62.5);

	// reg(ADDR) = 0
	*out++ = REG_ADDR | 0;
	*out++ = 0;
	// reg(COUNT) = delay
	*out++ = char(REG_COUNT | (delay >> 8));
	*out++ = char(delay);

	current[port].GPIOs[n] = pulse;
	current[port].UpdateCONFIG();
	*out++ = char(current[port].CONFIG1 >> 8);
	*out++ = char(current[port].CONFIG1);

	if (delay >= 0)
		*out++ = CMD_DELAY;

	current[port].GPIOs[n] = after;
	current[port].UpdateCONFIG();
	*out++ = char(current[port].CONFIG1 >> 8);
	*out++ = char(current[port].CONFIG1);

	Write(seq, out - seq);
	Flush_Commands();
	Sleep(ns / 1000);
}

void USBPod_I::Copy_TDI_Buffer_To_TDO
	(TAPPort port, unsigned offset, unsigned bytes)
{
	char doCopy[] =
		{ CMD_SETCURTAP | port,
		  REG_ADDR | (offset >> 8), offset,
		  REG_COUNT | (bytes >> 8), bytes,
		  CMD_COPY };
	Write(doCopy, sizeof(doCopy));
}

void USBPod_I::Do_Run_TCK(TAPPort port, bool onoff)
{
	unsigned& sr = current[port].CONFIG2;

	if (onoff)
		sr |= BIT_RUNTCK;
	else
		sr &= ~BIT_RUNTCK;

	char runtck[] =
		{
		CMD_SETCURTAP | port,
		sr >> 8, sr
		};

	Write(runtck, sizeof(runtck));
	Flush_Commands();
}

void USBPod_I::Pulse
	(CurrentTAP* now, unsigned X, unsigned ms)
{
	// number of outer delay loops per millisecond
	static const float DPERMS = (1e-3 / CLOCK_PERIOD) / 4096.0;

	if (ms > 1000) ms = 1000;
	unsigned delay = 4095 - unsigned(ms * DPERMS);

	char pulse[] =
		{
		CMD_SETCURTAP | (now - current),
		(now->CONFIG2 ^ X) >> 8, now->CONFIG2 ^ X,
		REG_ADDR | (delay >> 8) & 0xF, delay & 0xFF,
		REG_COUNT | 0, 0,
		CMD_DELAY,
		now->CONFIG2 >> 8, now->CONFIG2
		};

	Write(pulse, sizeof(pulse));
	Flush_Commands();
	Sleep(ms + 1);	// put this process to sleep for the duration
}

void USBPod_I::Set
	(CurrentTAP* now, unsigned X, bool on)
{
	unsigned& sr = now->CONFIG2;
	if (on)
		sr &= X;
	else
		sr |= X;

	char updateTAP[] =
		{ CMD_SETCURTAP | (now - current),
		  char(sr >> 8), char(sr) };

	Write(updateTAP, sizeof(updateTAP));
		Flush_Commands();
}

void USBPod_I::Set_TRST
	(TAPPort port, bool on)
{
	Set(&current[port], BIT_TRST, on);
}

void USBPod_I::Pulse_TRST
	(TAPPort port, unsigned ms)
{
	Pulse(&current[port], BIT_TRST, ms);
}

void USBPod_I::Set_RST
	(TAPPort port, bool on)
{
	Set(&current[port], BIT_RST, on);
}

void USBPod_I::Pulse_RST
	(TAPPort port, unsigned ms)
{
	Pulse(&current[port], BIT_RST, ms);
}

void USBPod_I::Set_DINT
	(TAPPort port, bool on)
{
	Set(&current[port], BIT_DINT, !on);
}

void USBPod_I::Pulse_DINT
	(TAPPort port, unsigned ms)
{
	Pulse(&current[port], BIT_DINT, ms);
}

}	// namespace JTAG

#if defined(__WIN32__)
#pragma argsused
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fwdreason, LPVOID lpvReserved)
{
	return 1;
}
#endif

//---------------------------------------------------------------------------

