//---------------------------------------------------------------------------

#include <iostream>
#include <fstream>
#include <ios>

#if defined(__WIN32__)
  #define WINDOWS_LEAN_AND_MEAN
  #include <windows.h>	// needed for FT types
  #pragma hdrstop
#endif

#include "FTD2XX.H"

using namespace std;

//---------------------------------------------------------------------------

FT_HANDLE ftHandle;
ifstream inf;

unsigned char* builtin = NULL;
unsigned builtin_length = 0;

unsigned char buf[16384], *out = buf;

inline int next()
{
	if (builtin)
		{
		if (!builtin_length) return EOF;
		builtin_length--;
		return *builtin++;
		}
	return inf.get();
}

inline void skip(int n)
{
	if (builtin)
		builtin += n;
	else
		inf.seekg(n, ios::cur);
}

void stop(int code)
{
	if (!FT_SUCCESS(FT_SetBitMode(ftHandle, 0x81, 0)))
		{
		cerr << "! Could not unset bit bang mode" << endl;
		}

	FT_Close(ftHandle);
	exit(code);
}

void flush()
{
	if (out != buf)
		{
		DWORD size = out - buf;
		DWORD wrote;
		if (!FT_SUCCESS(FT_Write(ftHandle, buf, size, &wrote)))
			{
			cerr << "! Failed write" << endl;
			stop(8);
			}

		if (wrote < size)
			{
			cerr << "! write was truncated" << endl;
			stop(9);
			}
		}
	out = buf;
}

void send_byte(char c)
{
	for (int n = 8; n--;)
		{
		out[1] = (out[0] = ((c >> 7) & 1)) | 0x80;
		out += 2;
		c <<= 1;
		}
	if (out == buf + sizeof(buf))
		flush();
}

#if defined(__WIN32__)
BOOL CALLBACK eproc(HANDLE, LPCTSTR, LPTSTR lpszName, LONG)
{
	cerr << "\t\t" << lpszName << endl;
	return TRUE;
}
#endif

void usage()
{
#if defined(__WIN32__)
	cerr << "usage: FPGAprog [<.bit file name>]|[=<test name>]" << endl;
	cerr << endl << "\tTest names are:" << endl;
	EnumResourceNames(NULL, "BITSTR", (ENUMRESNAMEPROC)eproc, 0);
#else
	cerr << "usage: FPGAprog <.bit file name>" << endl;
#endif
}

void pbin(unsigned v, unsigned w)
{
	while (w--)
		{
		cout << char('0'+ (v & 1));
		v >>= 1;
		}
}


#pragma argsused
int main(int argc, char* argv[])
{
#if defined(__WIN32__)
	if (argc > 2)
		{
#else
	if (argc != 2)
		{
#endif
		usage();
		return 1;
		}

	char* test = "jtag";

	if (argc > 1)
		if (argv[1][0] == '=')
			{
			test = &argv[1][1];
			}
		else
			{
			inf.open(argv[1], ios::in|ios::binary);
			if (!inf.is_open())
				{
				cerr << "! input .bit file not opened" << endl;
				usage();
				return 1;
				}
			test = NULL;
			}

#if defined(__WIN32__)
	if (test)
		{
		// get resource fork data
		HRSRC hR = FindResource(NULL, test, "BITSTR");
		builtin_length = SizeofResource(NULL, hR);
		builtin = (unsigned char*) LockResource(LoadResource(NULL, hR));
		if (!builtin)
			{
			cerr << "! built-in bit file " << test << "not found" << endl;
			usage();
			return 1;
			}
		}
#endif

	if (!FT_SUCCESS(FT_Open(0, &ftHandle)))
		{
		cerr << "! USB device not opened" << endl;
		return 2;
		}

	if (!FT_SUCCESS(FT_SetBaudRate(ftHandle, FT_BAUD_921600)))
		{
		cerr << "! " << endl;
		stop(2);
		}

	FT_SetTimeouts(ftHandle, 500, 500);

	if (!FT_SUCCESS(FT_SetBitMode(ftHandle, 0xFF, 1)))
		{
		cerr << "! Could not set bit bang mode" << endl;
		stop(3);
		}

	// skip headers
	int idx;
	int c1, c2;
	for (idx = 1; idx <= 3; idx++)
		{
		c1 = next();
		c2 = next();
		skip((c1 << 8) + c2);
		}

	// look for bitstream element
	for (;;)
		{
		c1 = next();
		if (c1 == 'e') break;
		c1 = next();
		c2 = next();
		skip((c1 << 8) + c2);
		if (c1 == EOF)
			{
			cerr << "! bad .bit file" << endl;
			stop(4);
			}
		}

	c1 = next();
	c2 = next();
	int length = ((c1 << 8) + c2) << 16;
	c1 = next();
	c2 = next();
	length |= (c1 << 8) + c2;

	cout << "Bitstream length = " << length << endl;

	// send it out!
	int count = 0;
	while (length--)
		{
		int c = next();
		if (c==EOF)
			{
			cerr << "! short .bit file" << endl;
			stop(4);
			}
		send_byte(char(c));
		if ((++count & 2047) == 0) cout << '.';
		}
	// FPGA needs some startup clocks
	send_byte(char(0));
	send_byte(char(0));
	send_byte(char(0));
	send_byte(char(0));
	flush();

	if (!FT_SUCCESS(FT_SetBitMode(ftHandle, 0xFF, 0)))
		{
		cerr << "! Could not unset bit bang mode" << endl;
		stop(5);
		}

	// wait for FPGA to come up and clock to lock
	DWORD x;
	for (;;)
		{
		cout << '+';
		FT_Purge(ftHandle, FT_PURGE_RX);
		x = 0;
		if (FT_SUCCESS(FT_Write(ftHandle, "\x10", 1, &x)))
			{
			char res;
			x = 0;
			if (FT_SUCCESS(FT_Read(ftHandle, &res, 1, &x)) &&
				(x == 1) &&
				(res & 0x02))
				break;
			}
		}

	// Get signature and version number
	
	// addr = 0
	FT_Write(ftHandle, "\xC0\x00", 2, &x);
	// count = 16
	FT_Write(ftHandle, "\x90\x10", 2, &x);
	// CMD_RDBUF
	FT_Write(ftHandle, "\x30", 1, &x);
	// retrieve the data
	x = 0;
	if (!FT_SUCCESS(FT_Read(ftHandle, buf, 16, &x)) || (x < 16))
		cout << endl << "! partial read of " << x << " bytes";
	else
		cout << endl << "Signature = \"" << buf << '"';

	FT_Purge(ftHandle, FT_PURGE_RX | FT_PURGE_TX);

	// Speed = 1MHz
	FT_Write(ftHandle, "\xF3\x0B", 2, &x);

	// addr = 0
	FT_Write(ftHandle, "\xC0\x00", 2, &x);
	// count = 2
	FT_Write(ftHandle, "\x90\x02", 2, &x);
	// CMD_WRBUF
	FT_Write(ftHandle, "\x20\x55\x55", 3, &x);
	cout << endl << "TDO: "; pbin(0x55, 8); pbin(0x55, 4);

	// BITS = 12
	FT_Write(ftHandle, "\xB0\x0C", 2, &x);
	// TMS = 101010101010
	FT_Write(ftHandle, "\xAA\xAA", 2, &x);
	cout << endl << "TMS: "; pbin(0xAA, 8); pbin(0xAA, 4);
	// and CMD_RUN!
	FT_Write(ftHandle, "\x43", 1, &x);

	Sleep(100);
	// addr = 0
	FT_Write(ftHandle, "\xC0\x00", 2, &x);
	// count = 2
	FT_Write(ftHandle, "\x90\x02", 2, &x);
	// CMD_RDBUF
	FT_Write(ftHandle, "\x30", 1, &x);
	FT_Read(ftHandle, buf, 2, &x);

	cout << endl << "Rcv: "; pbin(buf[0], 8); pbin(buf[1], 4);

	// clear FPGA
	FT_Write(ftHandle, "\x89\xFF", 2, &x);

	FT_Close(ftHandle);

	return 0;
}

//---------------------------------------------------------------------------
